/*
 * Copyright 2002 Yrj Kari-Koskinen and Henri Sivonen.
 *
 * The contents of this file are subject to the Artistic License.
 * You may not use this file except in compliance with the Artistic
 * License (which you should have received along with this file).
 */

package dualroids;

import java.awt.event.*;
import java.io.IOException;

/**
 * Class for handling keyboard input and <code>windowClosing</code>
 * window event.
 *
 * @author Henri Sivonen, hsivonen@iki.fi
 * @author Yrj Kari-Koskinen, yrjo.kari-koskinen@hut.fi
 */
public final class InputHandler extends WindowAdapter implements KeyListener {

    /**
     * The ship that is controlled by this <code>InputHandler</code>.
     */
    SpaceShip ship;

    /**
     * The associated <code>ClientModelConnector</code>.
     */
    ClientModelConnector connector;

    /**
     * Contructs a new <code>InputHandler</code>.
     *
     * @param connector the <code>ClientModelConnector</code> to
     *                  be associated with this <code>InputHandler</code>
     */
    public InputHandler(ClientModelConnector connector) {
        this.connector = connector;
        this.ship = null;
    }

    /**
     * Registers a <code>SpaceShip</code> for contrilling through
     * this <code>InputHandler</code>.
     *
     * @param ship the ship to be controlled
     */
    public void registerShip(SpaceShip ship) {
        this.ship = ship;
    }

    /**
     * Handles key pressed events.
     *
     * @param ke the event
     */
    public void keyPressed(KeyEvent ke) {
        try {
            if(this.ship != null) {
                int key = ke.getKeyCode();
                switch(key) {
                    case KeyEvent.VK_UP:
                        this.ship.startThrust(this.connector.
                                              currentServerMillis());
                        break;
                    case KeyEvent.VK_LEFT:
                        this.ship.startTurnCCW(this.connector.
                                               currentServerMillis());
                        break;
                    case KeyEvent.VK_RIGHT:
                        this.ship.startTurnCW(this.connector.
                                              currentServerMillis());
                        break;
                }
            }
        } catch(NullPointerException npe) {
            return;
        }
    }

    /**
     * Handles key released events.
     *
     * @param ke the event
     */
    public void keyReleased(KeyEvent ke) {
        try {
            if(this.ship != null) {
                int key = ke.getKeyCode();
                switch(key) {
                    case KeyEvent.VK_SPACE:
                        this.ship.fireBullet(this.connector.
                                             currentServerMillis());
                        break;
                    case KeyEvent.VK_UP:
                        this.ship.stopThrust(this.connector.
                                             currentServerMillis());
                        break;
                    case KeyEvent.VK_LEFT:
                        this.ship.stopTurnCCW(this.connector.
                                              currentServerMillis());
                        break;
                    case KeyEvent.VK_RIGHT:
                        this.ship.stopTurnCW(this.connector.
                                             currentServerMillis());
                        break;
                    case KeyEvent.VK_Q:
                        this.windowClosing(null);
                        break;
                }
            }
        } catch(NullPointerException npe) {
            return;
        }
    }

    /**
     * Dummy handler for key typed events.
     *
     * @param ke the event
     */
    public void keyTyped(KeyEvent ke) {}


    /**
     * Clean up the window and initiate a network close procedure when the
     * window close event is made. The network close will eventually cause
     * an <code>exit()</code> to be called when everything is safely closed.
     *
     * @param event the window event
     */
    public void windowClosing(WindowEvent event) {
        // clean up the GameFrame window
        GameFrame gf = GameFrame.getInstance();
        if (gf != null && gf.isShowing()) {
            gf.dispose();
        }

        try {
            // send close via cmc
            this.connector.close();
        } catch(IOException ioe) {
            // we are not interested in this exception at this point
        }
    }
}
