/*
 * Copyright 2002 Henri Sivonen.
 *
 * The contents of this file are subject to the Artistic License.
 * You may not use this file except in compliance with the Artistic
 * License (which you should have received along with this file).
 */

package dualroids;

import java.awt.*;
import java.awt.geom.*;
import java.io.IOException;

/**
 * The representation of a space ship.
 *
 * <h3>A Note About Thread Safety</h3>
 * <p>The implementation of <code>SpaceShip</code> is thread-safe
 * in the sense that the one thread may call the setters
 * while another calls the getters. However, there may
 * <strong>not</strong> be multiple threads calling
 * <code>getTransformedShape</code> and <code>getVertices</code>.
 * They are designed to be used from a single thread only.</p>
 * <p>Also, that thread should not retain references to
 * objects returned by <code>getTransformedShape</code> and
 * <code>getVertices</code> across the method calls.
 * When the methods are called the objects they returned
 * previously will no longer be valid.
 * </p>
 *
 * @author Henri Sivonen, hsivonen@iki.fi
 */
public final class SpaceShip extends SpatialObject {

    /**
     * A constant indicating the number of vertices in
     * the ship polygon.
     */
    private static final int NUM_OF_VERTICES = 3;

    /**
     * The number of bullets this ship can have flying at
     * a time.
     */
    private static final int NUM_OF_BULLETS = 6;

    /**
     * A constant for indicating that the ship is not turning.
     */
    private static final int TURN_NO = 0;

    /**
     * A constant for indicating that the ship is turning counter-clockwise.
     */
    private static final int TURN_CCW = -1;

    /**
     * A constant for indicating that the ship is turning clockwise.
     */
    private static final int TURN_CW = 1;

    /**
     * The angular velocity of the ship when the ship is turning.
     */
    private static final double ANGULAR_VELOCITY = 0.003;

    /**
     * The angular velocity of ship debris.
     */
    private static final double DEBRIS_ANGULAR_VELOCITY = 0.005;

    /**
     * The acceleration of the ship when the ship is thrusting.
     */
    private static final double ACCELERATION = 0.00005;

    /**
     * The maximum velocity of the ship.
     */
    private static final double MAX_VELOCITY = 0.085;

    /**
     * The minimum delay between firing bullets.
     */
    private static final long BULLET_INTERVAL = 600L;

    /**
     * A factor used when calculating the velocity of debris.
     */
    private static final float DEBRIS_VELOCITY_FACTOR = 0.002f;
    
    /** The amount of time used for one outline color in explosion */
    private static final float COLOR_STEP =
        Server.NEW_SHIP_DELAY / SpatialObject.COLOR_TABLE.length;

    /**
     * An array holding coordinate constants used when drawing debris.
     */
    private static final float[] DEBRIS_POINTS = {
        10.0f,  // start0X
         2.5f,  // start0Y
       -10.0f,  // end0X
        -2.5f,  // end0Y
         0.0f,  // start1X
        -5.0f,  // start1Y
         0.0f,  // end1X
         5.0f,  // end1Y
       -10.0f,  // start2X
         2.5f,  // start2Y
        10.0f,  // end2X
        -2.5f   // end2Y
    };

    /**
     * The coordinated of the center poins of the edges of the ship.
     */
    private static final float[] CENTER_POINTS = {
         0.0f,  // center0X
        -2.5f,  // center0Y
       -10.0f,  // center1X
         0.0f,  // center1Y
         0.0f,  // center2X
         2.5f   // center2Y
    };
    
    /**
     * The rotation angle (in radians) of the spatial object at the
     * moment indicated by <code>timestamp</code>.
     */
    private double angle;

    /**
     * An array holding the untranslated vertices of the
     * ship polygon.
     */
    private transient Point2D[] vertices;

    /**
     * An array holding the translated copies of the
     * vertices ship polygon.
     */
    private transient Point2D[] translatableVertices;

    /**
     * An array holding all the (recyclable) bullets
     * owned by this ship.
     */
    private transient Bullet[] bullets;

    /**
     * The index of the previously fired bullet. Used in bullet
     * recycling rotation.
     */
    private transient int bulletFireIndex;

    /**
     * Indicates when the next bullet can fired at the earliest.
     */
    private transient long earliestNextBullet;

    /**
     * The <code>ClientModelConnector</code> through
     * which this ship can send its state to the server.
     */
    private transient ClientModelConnector connector;

    /**
     * Indicates when a 'new' ship appears after the ship has exploded.
     */
    private transient long newAppears;

    /**
     * Indicates whether the ship is thrusting.
     */
    private transient boolean thrustOn;

    /**
     * Indicates turning state of the ship.
     */
    private transient int turning;

    /** 
     * the outline color as an index to 
     * <code>SpatialObject.COLOR_TABLE</code> 
     */
    private transient int outlineColor;

    /**
     * A helper method for updating the transormation.
     *
     * @param time the point of time for which the transformation
     *             is needed.
     */
    private final void updateTransformation(long time) {
        double deltaT, x, y, cosAngle, sinAngle, theta;
        synchronized(this) {
            deltaT = (double)(time - this.timestamp);
            x = this.locationX + this.velocityX * deltaT;
            y = this.locationY + this.velocityY * deltaT;
            theta = this.angle;
        }
        x = moveToRangeX(x);
        y = moveToRangeY(y);
        cosAngle = Math.cos(theta);
        sinAngle = Math.sin(theta);
        this.transform.setTransform(cosAngle,
                                    sinAngle,
                                    -sinAngle,
                                    cosAngle,
                                    x,
                                    y);
    }

    /**
     * Returns the vertices of the ship polygon at <code>time</code>.
     * The caller should not write to the returned array.
     *
     * @param time the point of time for which the vertices
     *             are needed.
     * @return an array of the vertices of the ship polygon
     */
    public final Point2D[] getVertices(long time) {
        if(this.newAppears <= time) {
            this.updateTransformation(time);
            this.transform.transform(this.vertices,
                                     0,
                                     this.translatableVertices,
                                     0,
                                     this.vertices.length);
            return this.translatableVertices;
        } else {
            return null;
        }
    }

    /**
     * Returns a properly transformed <code>Shape</code> representing
     * the spatial object at <code>time</code>
     *
     * @param time the moment in game (server) time for which the
     *             representation is requested
     * @return the <code>Shape</code> representation at <code>time</code>
     *         or null if the object isn't visible
     */
    public final Shape getTransformedShape(long time) {
        if(this.newAppears <= time) {
	    if (time < (this.newAppears + Server.INVULNERABILITY_PERIOD)) {
		this.outlineColor = SpatialObject.COLOR_TABLE_GRAY_50;
	    } else {
		this.outlineColor = SpatialObject.COLOR_TABLE_WHITE;
	    }

            this.updateTransformation(time);
            return this.geometry.createTransformedShape(this.transform);
        } else {
            double deltaT, effDebrisAngle;
            float cosShipAngle, sinShipAngle, cosEffDebrisAngle,
                  sinEffDebrisAngle, x, y, deltaTf, startX, startY, endX, endY,
                  centerX, centerY, rCenterX, rCenterY, rStartX, rEndX,
                  rStartY, rEndY;
            deltaT = (double)(time - this.newAppears + Server.NEW_SHIP_DELAY);

            // calculate the color from the time delta
            this.outlineColor =  16 - (int)(deltaT / COLOR_STEP);
            if(this.outlineColor < 0) {
                this.outlineColor = 0;
            }
            if(this.outlineColor > 16) {
                this.outlineColor = 16;
            }


            deltaTf = (float)deltaT;
            x = (float)this.locationX;
            y = (float)this.locationY;
            GeneralPath debrisGeometry = 
                new GeneralPath(GeneralPath.WIND_NON_ZERO,
                                NUM_OF_VERTICES + 2);
            cosShipAngle = (float)Math.cos(this.angle);
            sinShipAngle = (float)Math.sin(this.angle);
            cosEffDebrisAngle = (float)Math.cos(this.angle
                                                + DEBRIS_ANGULAR_VELOCITY
                                                * deltaT);
            sinEffDebrisAngle = (float)Math.sin(this.angle
                                                + DEBRIS_ANGULAR_VELOCITY
                                                * deltaT);
            int index;
            for(int i = 0; i < NUM_OF_VERTICES - 1; i++) {
                index = i * 4;
                startX = DEBRIS_POINTS[index];
                startY = DEBRIS_POINTS[index + 1];
                endX = DEBRIS_POINTS[index + 2];
                endY = DEBRIS_POINTS[index + 3];
                centerX = CENTER_POINTS[i * 2];
                centerY = CENTER_POINTS[i * 2 + 1];

                // rotate
                rStartX = startX * cosEffDebrisAngle - 
                    startY * sinEffDebrisAngle;
                rEndX = endX * cosEffDebrisAngle - endY * sinEffDebrisAngle;
                rStartY = startX * sinEffDebrisAngle + 
                    startY * cosEffDebrisAngle;
                rEndY = endX * sinEffDebrisAngle + endY * cosEffDebrisAngle;

                rCenterX = centerX * cosShipAngle - centerY * sinShipAngle;
                rCenterY = centerX * sinShipAngle + centerY * cosShipAngle;
                
                // translate
                float deltaX = DEBRIS_VELOCITY_FACTOR * rCenterX * 
                    deltaTf + x + rCenterX;
                float deltaY = DEBRIS_VELOCITY_FACTOR * rCenterY * 
                    deltaTf + y + rCenterY;
                
                startX = rStartX + deltaX;
                endX = rEndX + deltaX;
                startY = rStartY + deltaY;
                endY = rEndY + deltaY;

                // draw
                debrisGeometry.moveTo(startX, startY);
                debrisGeometry.lineTo(endX, endY);
            }
            return debrisGeometry;            
        }
    }

    /**
     * Contructs a new <code>SpaceShip</code>
     */
    public SpaceShip(double locationX,
                     double locationY,
                     long timestamp)
    {
        super(locationX,
              locationY,
              0.0,
              0.0,
              timestamp);
        this.angle = Math.PI / 2.0;
        this.init();
    }

    /**
     * Does real ship initialization. This method is separate from
     * the constuctur due to problems with object serialization.
     */
    private void init() {
        this.vertices = new Point2D[NUM_OF_VERTICES];
        this.vertices[0] = new Point2D.Double(10.0, 0.0);
        this.vertices[1] = new Point2D.Double(-10.0, -5.0);
        this.vertices[2] = new Point2D.Double(-10.0, 5.0);
        this.geometry = new GeneralPath(GeneralPath.WIND_NON_ZERO,
                                        NUM_OF_VERTICES + 2);
        this.geometry.moveTo(10.0f, 0.0f);
        this.geometry.lineTo(-10.0f, -5.0f);
        this.geometry.lineTo(-10.0f, 5.0f);
        this.geometry.closePath();
        this.translatableVertices = new Point2D[NUM_OF_VERTICES];
        for(int i = 0; i < NUM_OF_VERTICES; i++) {
            this.translatableVertices[i] = new Point2D.Double();
        }        
        this.bullets = new Bullet[NUM_OF_BULLETS];
        for(int i = 0; i < NUM_OF_BULLETS; i++) {
            this.bullets[i] = new Bullet();
        }
        this.bulletFireIndex = 0;
        this.connector = null;
        this.transform = new AffineTransform();
        this.newAppears = -1L;
        this.thrustOn = false;
        this.turning = TURN_NO;
        this.earliestNextBullet = -1L;
    }
    
    /**
     * Fires a bullet and marks it fired at <code>time</code>.
     * @param time when the bullet was fired
     */
    public final void fireBullet(long time) {
        if((this.earliestNextBullet > time) ||
           (this.newAppears > time)) {
            return;
        }
        int i = this.bulletFireIndex;
        while(!this.bullets[i].readyToBeFired()) {
            i++;
            i %= NUM_OF_BULLETS;
            if(i == this.bulletFireIndex) {
                return; // all bullets used up
            }
        }
        this.bulletFireIndex = i;
        double deltaT, x, y, shipVelocityX, shipVelocityY, shipAngle;
        synchronized(this) {
            deltaT = (double)(time - this.timestamp);
            x = this.locationX + this.velocityX * deltaT;
            y = this.locationY + this.velocityY * deltaT;
            shipVelocityX = this.velocityX;
            shipVelocityY = this.velocityY;
            shipAngle = this.angle;
        }
        x = moveToRangeX(x);
        y = moveToRangeY(y);
        this.bullets[i].fire(x,
                             y,
                             shipVelocityX,
                             shipVelocityY,
                             shipAngle,
                             time);
        this.bullets[i].writeToConnector(i, this.connector);
        this.earliestNextBullet = time + BULLET_INTERVAL;
    }

    /**
     * Registers a <code>ClientModelConnector</code> with the ship for
     * sending ship state to the server. Used only for the ship
     * controlled by the client itself.
     *
     * @param connector the <code>ClientModelConnector</code> to be
     *                  registered
     */
    public final void registerConnector(ClientModelConnector connector) {
        this.connector = connector;
    }

    /**
     * Returns the <code>Bullet</code> array of this ship. The caller
     * should not write to the array.
     *
     * @return the <code>Bullet</code> array of this ship
     */
    public final Bullet[] getBullets() {
        return this.bullets;
    }

    /**
     * Returns the outline color of the ship.
     * @return the outline color of the ship
     */
    public final Color getOutlineColor() {
        return SpatialObject.COLOR_TABLE[this.outlineColor];
    }

    /**
     * Makes the ship explode.
     *
     * @param newAppears when the new ship appears
     */
    public final void explode(long newAppears) {
        synchronized(this) {
            this.newAppears = newAppears;
            this.velocityX = 0.0;
            this.velocityY = 0.0;
            this.thrustOn = false;
            this.turning = TURN_NO;
        }
    }


    /**
     * Updates the state of this ship to the values
     * corresponding to a moment in time given as the argument. The
     * timestamp is updated accordingly.
     *
     * @param time the new timestamp
     */
    public final void updateState(long time) {
        double deltaT, x, y, absVelocity;
        synchronized(this) {
            if(this.newAppears <= time) {
                deltaT = (double)(time - this.timestamp);
                this.locationX = moveToRangeX(this.locationX
                                              + this.velocityX
                                              * deltaT);
                this.locationY = moveToRangeY(this.locationY
                                              + this.velocityY
                                              * deltaT);
                if(this.turning == TURN_CW) {
                    this.angle += ANGULAR_VELOCITY * deltaT;
                    while(this.angle < 0.0) {
                        this.angle += 2.0 * Math.PI;
                    }
                    this.angle %= 2.0 * Math.PI;
                } else if(this.turning == TURN_CCW) {
                    this.angle -= ANGULAR_VELOCITY * deltaT;
                    while(this.angle < 0.0) {
                        this.angle += 2.0 * Math.PI;
                    }
                    this.angle %= 2.0 * Math.PI;
                }
                if(this.thrustOn) {
                    this.velocityX += ACCELERATION * deltaT * Math.cos(this.angle);
                    this.velocityY += ACCELERATION * deltaT * Math.sin(this.angle);
                    absVelocity = Math.sqrt(this.velocityX * this.velocityX
                                            + this.velocityY * this.velocityY);
                    if(absVelocity > MAX_VELOCITY) {
                        this.velocityX *= MAX_VELOCITY / absVelocity;
                        this.velocityY *= MAX_VELOCITY / absVelocity;
                    }
                } 
            }
            this.timestamp = time;
        }
    }

    /**
     * Starts thrusting.
     * @param time when the thrusting was started
     */
    public final void startThrust(long time) {
        this.updateState(time);
        synchronized(this) {
            this.thrustOn = true;
        }
    }

    /**
     * Stops thrusting.
     * @param time when the thrusting was stopped
     */    
    public final void stopThrust(long time) {
        this.updateState(time);
        synchronized(this) {
            this.thrustOn = false;
        }
    }

    /**
     * Starts turning counter-clockwise.
     * @param time when the turning was started
     */
    public final void startTurnCCW(long time){
        this.updateState(time);
        synchronized(this) {
            this.turning = TURN_CCW;
        }
    }

    /**
     * Stops turning counter-clockwise.
     * @param time when the turning was stopped
     */
    public final void stopTurnCCW(long time){
        this.updateState(time);
        synchronized(this) {
            if(this.turning == TURN_CCW) {
                this.turning = TURN_NO;
            }
        }
    }

    /**
     * Starts turning clockwise.
     * @param time when the turning was started
     */
    public final void startTurnCW(long time){
        this.updateState(time);
        synchronized(this) {
            this.turning = TURN_CW;
        }
    }

    /**
     * Stops turning clockwise.
     * @param time when the turning was stopped
     */
    public final void stopTurnCW(long time){
        this.updateState(time);
        synchronized(this) {
            if(this.turning == TURN_CW) {
                this.turning = TURN_NO;
            }
        }
    }

    /**
     * Sets the state of the ship to values received from the network.
     *
     * @param locationX the x coordinate of the space ship
     * @param locationY the y coordinate of the space ship
     * @param velocityX the x component of the velocity of the space ship
     * @param velocityY the y component of the velocity of the space ship
     * @param angle the rotational angle of the space ship
     * @param timestamp the timestamp
     */
    public final void setStateFromRemote(double locationX,
                                         double locationY,
                                         double velocityX,
                                         double velocityY,
                                         double angle,
                                         long timestamp)
    {
        synchronized(this) {
            this.locationX = locationX;
            this.locationY = locationY;
            this.velocityX = velocityX;
            this.velocityY = velocityY;
            this.angle = angle;
            this.timestamp = timestamp;
        }
    }

    /**
     * Writes the state of ship to the registered
     * <code>ClientModelConnector</code>
     */
    public final void writeToConnector() {
        synchronized(this) {
            try {
                this.connector.writeShipState(this.locationX,
                                              this.locationY,
                                              this.velocityX,
                                              this.velocityY,
                                              this.angle,
                                              this.timestamp);
            } catch(IOException ioe) {
                Runner.fatalError("Error in network connection!");
            }
        }
    }

    /**
     * Serializes the ship for sending to the clients.
     *
     * @param out the target stream
     */
    private void writeObject(java.io.ObjectOutputStream out)
        throws IOException
    {
        out.writeLong(this.timestamp);
        out.writeDouble(this.locationX);
        out.writeDouble(this.locationY);
        out.writeDouble(this.velocityX);
        out.writeDouble(this.velocityY);
        out.writeDouble(this.angle);

    }

    /**
     * Deserializes the ship on the client side.
     *
     * @param in the source stream
     */
    private void readObject(java.io.ObjectInputStream in)
        throws IOException, ClassNotFoundException
    {
        this.timestamp = in.readLong();
        this.locationX = in.readDouble();
        this.locationY = in.readDouble();
        this.velocityX = in.readDouble();
        this.velocityY = in.readDouble();
        this.angle = in.readDouble();
        this.init();
    }  
}
